// wuiobjdb.c - implement WUIMAN object database.
// Copyright (c) 1993 by Ron Burk

#include <ctype.h>
#include <errno.h>
#include <limits.h>
#include <string.h>
#include <stdio.h>      // sscanf()

#include "wuidbint.h"
#include "wuierror.h"
#include "wuistd.h"
#include "wuiobjdb.h"


////////////////////////////////////////////////////////////
// TWuiObjects - Provide arrays of AWuiObject pointers

// TWuiObjects - initialize object with desired array size.
TWuiObjects::TWuiObjects(int Size)
    :   TWuiGenericList(Size)
    {   MEMBERASSERT();
    }

// (private)
// Find - Return position of named child

int     TWuiObjects::Find(const TWuiName Which)
    {   MEMBERASSERT();

    for(int iChild = 0; iChild < NObjects(); ++iChild)
        {
        ASSERT(Get(iChild) != NULL);
        if(Get(iChild)->Name() == Which)
            return iChild;
        }
    return -1;
    }

AWuiObject *TWuiObjects::Insert(AWuiObject *NewObject, const TWuiName To)
    {   MEMBERASSERT();
    ASSERT(NewObject != 0);
    return (AWuiObject *)Insert(NewObject, Find(To));
    }

AWuiObject *TWuiObjects::Insert(AWuiObject *NewObject, int Position)
    {   MEMBERASSERT();
    ASSERT(NewObject != 0);
    return (AWuiObject *)TWuiGenericList::Insert((void *)NewObject, Position);
    }

AWuiObject  *TWuiObjects::Get(const TWuiName Which)
    {   MEMBERASSERT();
    int     Position = Find(Which);
    if(Position >= 0)
        return Get(Position);
    else
        return 0;
    }

AWuiObject *TWuiObjects::Remove(const TWuiName Which)
    {   MEMBERASSERT();
    int     Position = Find(Which);
    if(Position >= 0)
        return Remove(Position);
    else
        return 0;
    }

////////////////////////////

//static TWuiRegister RegisterObject(
//    new TWuiObject(C_TOBJECT));

// TWuiObject - An AWuiObject that supports attributes.
TWuiObject::TWuiObject(const TWuiName ObjectName,
                              int NAttributes)
    :   AWuiObject(ObjectName), Attributes_(0)
    {   MEMBERASSERT();
    // note that it is Ok if NAttributes is 0,
    // as the goal here is just to create the array.
    Attributes_ = new TWuiObjects(NAttributes+3);
    // All objects support these methods
    DefineMethod(M_OBJECTISA);
    DefineMethod(M_OBJECTDELETE);
    DefineMethod(M_OBJECTRENAME);
    DefineMethod(M_OBJECTCLONE);
    DefineMethod(M_OBJECTDUMP);
    DefineMethod(M_OBJECTCLASS);
    DefineMethod(M_OBJECTPARENT);
    DefineMethod(M_OBJECTCHILD);
    DefineMethod(M_OBJECTATTRIBUTE);
    // All objects support this property
    ASSERT(P_Persistent != NULL);
    Attributes_->Insert(P_Persistent);
    }


// Set - Handle modifications to "Persistent" property.
//
// 1) making an object persistent makes all its ancestor
//    containers persistent.
// 2) making a container non-persistent makes all its child
//    objects (and their attributes) non-persistent.

long    TWuiObject::Set(TWuiPath Path, TWuiArgs *Args,
                                         const char *Buffer)
    {   MEMBERASSERT();
    if(AWuiObject::TraceFlag)
        Trace("TWuiObject::Set", Path, Args);
    long    Result = AWuiObject::Set(Path, Args, Buffer);
    // if not attribute "p_Persistent", then we don't care
    if(Args->AttributeName() != P_PERSISTENT)
        return AWuiObject::Set(Path, Args, Buffer);
    // if set was not successful, then we don't care
    if(Result != 1 && Result != 0)
        return Result;
    if(Result == 1)  // if enabling persistence
        SetFlags(PERSISTENT);
    // else if we are being made non-persistent
    else if(!Path.Descend())
        {
        ClearFlags(PERSISTENT);
        Path.Ascend();
        TWuiObjects *List = Attributes();
        if(List)
            Broadcast(List, Path, Args, Buffer, 0, TRUE);
        List    = Children();
        if(List)
            Broadcast(List, Path, Args, Buffer, 0, TRUE);
        AWuiObject::Database->DeletePath(Path.AbsolutePath());
        }

    return Result;
    }

TWuiName    TWuiObject::ClassName() const
    {   MEMBERASSERT();
    return C_TOBJECT;
    }
TWuiName    TWuiObject::ParentClassName() const
    {   MEMBERASSERT();
    return C_ROOT;
    }

void    TWuiObject::Copy(const AWuiObject *Other,
                                  int Depth)
    {   MEMBERASSERT();
    AWuiObject::Copy(Other, Depth);
    }

AWuiObject  *TWuiObject::Clone(const TWuiName NewName,
                                  int Depth)
    {   MEMBERASSERT();
    ASSERT(Attributes_ != NULL);
    AWuiObject *Result = new TWuiObject(NewName,
                                  Attributes()->NObjects());
    if(Result == NULL)
        return Result;
    else
        {
        Result->Copy(this, Depth);
        return Result;
        }
    }
long    TWuiObject::Save(TWuiPath Path, int Depth) const
    {   MEMBERASSERT();
    char    MyClass[TWuiName::MAXNAME+1];
    ClassName().GetName(MyClass);
    char    MyName[TWuiName::MAXNAME+1];
    Name().GetName(MyName);
    char    MyPath[TWuiPath::MAXPATH];
    Path.Ascend();
    Path.CopyRelativePath(MyPath);
    Path.Descend();
DEBUG_TRACE(">TWuiObject::Save< name='%s', class='%s', Path='%s' '%s', %d",
            MyName, MyClass, Path.AbsolutePath(), MyPath, Depth);

    if(MyPath[0] != '\0' && !Database->WriteItem(MyPath,
                        MyName, MyClass))
        {
        DEBUG_ERROR("Write of '%s':'%s'='%s' failed",
            Path.AbsolutePath(), MyName, MyClass);
        return WUIMAN_ERR_WRITE_FAILED;
        }
    else
        // save my attributes and any children to
        // desired depth
        return AWuiObject::Save(Path, Depth);
    }


TWuiObjects *TWuiObject::Attributes() const
    {   MEMBERASSERT();
    return Attributes_;
    }

int     TWuiObject::IsA(const TWuiName Class,
                                  const TWuiName Method) const
    {   MEMBERASSERT();
    if(Class == C_TOBJECT)
        {
        if(Method == "")
            return TRUE;
        else if(GetAttribute(Method) != NULL)
            return TRUE;
        }
    return FALSE;
    }

long    TWuiObject::DefineAttribute(AWuiAttribute *Attr)
    {   MEMBERASSERT();
    if(Attr == NULL)
        {
        DEBUG_ERROR("DefineAttribute(): NULL attribute "
                               "value (assume new failed)");
        return WUIMAN_ERR_NEW_FAILED;
        }
    TWuiObjects *List = Attributes();
    if(List == NULL)
        {
        DEBUG_ERROR("DefineAttribute(): This object does "
            "not support attributes.");
        return WUIMAN_ERR_TARGET_HATES_ATTR;
        }
    if(List->Get(Attr->Name()))
        {
        char    Name[TWuiName::MAXNAME];
        DEBUG_ERROR("DefineAttribute('%s'): Name collision.",
                                 Attr->Name().GetName(Name));
        DumpObject();
        return WUIMAN_ERR_DUPLICATE_NAME;
        }
    List->Insert(Attr);
    return TRUE;
    }


////////////


// TWuiNonContainer - An AWuiObject that supports attributes.
TWuiNonContainer::TWuiNonContainer(const TWuiName ObjectName,
                              int NAttributes)
    :   TWuiObject(ObjectName, NAttributes)
    {   MEMBERASSERT();
    }

// Class constructor -- just return master object.
AWuiObject *TWuiNonContainer::ClassConstructor(
                                        const TWuiName Name)
    {
    return new TWuiNonContainer(Name);
    }

TWuiName    TWuiNonContainer::ClassName() const
    {   MEMBERASSERT();
    return C_TNONCONTAINER;
    }
TWuiName    TWuiNonContainer::ParentClassName() const
    {   MEMBERASSERT();
    return C_TOBJECT;
    }

void    TWuiNonContainer::Copy(const AWuiObject *Other, int Depth)
    {   MEMBERASSERT();
    TWuiObject::Copy(Other, Depth);
    }

AWuiObject  *TWuiNonContainer::Clone(const TWuiName NewName, int Depth)
    {   MEMBERASSERT();
    ASSERT(Children() == NULL);
    ASSERT(Attributes() != NULL);
    AWuiObject *Result = new TWuiNonContainer(NewName,
            Attributes()->NObjects());
    if(Result == NULL)
        return Result;
    else
        {
        Result->Copy(this, Depth);
        return Result;
        }
    }


int     TWuiNonContainer::IsA(const TWuiName Class, const TWuiName Method) const
    {   MEMBERASSERT();
    if(Class == C_TNONCONTAINER)
        {
        if(Method == "")
            return TRUE;
        else if(GetAttribute(Method) != NULL)
            return TRUE;
        }
    return TWuiObject::IsA(Class, Method);
    }



////////////
// 11/06/93

AWuiObject *TWuiContainer::ClassConstructor(const TWuiName Name)
    {
    return new TWuiContainer(Name);
    }

// TWuiContainer - A TWuiObject that supports children.
TWuiContainer::TWuiContainer(const TWuiName ObjectName,
                              int NAttributes, int NChildren)
    :   TWuiObject(ObjectName, NAttributes+1), Children_(0)
    {   MEMBERASSERT();
    // note that it is Ok if NChildren or NAttributes is 0,
    // as the goal here is just to create the arrays.
    Children_ = new TWuiObjects(NChildren);
    }

// Set - Handle modifications to "Persistent" property.
//
// 1) making an object persistent makes all its ancestor
//    containers persistent.
// 2) making a container non-persistent makes all its child
//    objects (and their attributes) non-persistent.

long    TWuiContainer::Set(TWuiPath Path, TWuiArgs *Args,
                                         const char *Buffer)
    {   MEMBERASSERT();
#if 0
    // if not attribute "p_Persistent", then we don't care
    if(Args->AttributeName() != P_PERSISTENT)
        return Result;
    // if not setting or clearing, then we don't care
    if(Buffer == NULL || Buffer[0] == '\0')
        return Result;
    // determine whether enabling or disabling persistence
    int TurnOn = (atoi(Buffer) != 0);

    if(TurnOn)  // if enabling persistence
        {
        if(!Path.Descend()) // if message was for me
            return Result;  // then I am already set
        else    // else, make me persistent too!
            {
            Path.Ascend();
            return AWuiObject::Set(Path, Args, Buffer);
            }
        }
    // else if we are being made non-persistent
    else if(!Path.Descend())
        {
        Path.Ascend();
        TWuiObjects *List = Attributes();
        if(List)
            
        List    = Children();
        if(List)
            ;
        }
    else
#endif
    long    Result = AWuiObject::Set(Path, Args, Buffer);
        return Result;
    }

TWuiName    TWuiContainer::ClassName() const
    {   MEMBERASSERT();
    return C_TCONTAINER;
    }
TWuiName    TWuiContainer::ParentClassName() const
    {   MEMBERASSERT();
    return C_TOBJECT;
    }

void    TWuiContainer::Copy(const AWuiObject *Other, int Depth)
    {   MEMBERASSERT();
    TWuiObject::Copy(Other, Depth);
    }

AWuiObject  *TWuiContainer::Clone(const TWuiName NewName, int Depth)
    {   MEMBERASSERT();
    ASSERT(Children() != NULL);
    ASSERT(Attributes() != NULL);
    AWuiObject *Result = new TWuiContainer(NewName,
            Attributes()->NObjects(), Children()->NObjects());
    if(Result == NULL)
        return Result;
    else
        {
        Result->Copy(this, Depth);
        return Result;
        }
    }
TWuiObjects *TWuiContainer::Children() const
    {   MEMBERASSERT();
    return Children_;
    }


int     TWuiContainer::IsA(const TWuiName Class, const TWuiName Method) const
    {   MEMBERASSERT();
    if(Class == C_TCONTAINER)
        {
        if(Method == "")
            return TRUE;
        else if(GetAttribute(Method) != NULL)
            return TRUE;
        }
    return TWuiObject::IsA(Class, Method);
    }
////////////////

const char TWuiOrderedContainer::SEPARATOR = ';';


AWuiObject *TWuiOrderedContainer::ClassConstructor(const TWuiName Name)
    {
    return new TWuiOrderedContainer(Name);
    }

// TWuiOrderedContainer - A TWuiObject that supports children.
TWuiOrderedContainer::TWuiOrderedContainer(const
        TWuiName ObjectName, int NAttributes, int NChildren)
    :   TWuiContainer(ObjectName, NAttributes+1, NChildren),
        PropertyNames(NULL), MruList(NULL), ObjectOrder(NULL)
    {   MEMBERASSERT();
    DefineAttribute(new TWuiStringProperty(P_OBJECTORDER, &ObjectOrder));
    }

TWuiOrderedContainer::~TWuiOrderedContainer()
    {   MEMBERASSERT();
    if(PropertyNames != NULL)
        delete[] PropertyNames;
    if(MruList != NULL)
        delete MruList;
    }

// Load - Enforce ordered relations at load time.
long    TWuiOrderedContainer::Load(TWuiPath Path,
                                          const char *Value)
    {   MEMBERASSERT();
DEBUG_TRACE("TWuiOrderedContainer::Load(Path='%s','%s')",
        Path.AbsolutePath(), Path.RelativePath());
    long    Result = TWuiContainer::Load(Path, Value);
    char    Buffer[TWuiPath::MAXPATH];
DEBUG_TRACE("[]<>TWuiOrderedContainer::Load(Path='%s','%s')",
        Path.AbsolutePath(), Path.RelativePath());
    NewObjectOrder(Path);
    return Result;
    }

TWuiName   *TWuiOrderedContainer::ParseOrderString(
                             const char *String, int *Total)
    {   MEMBERASSERT();
DEBUG_TRACE("--->ParseOrderString('%s')", String);
    TWuiName    *Result = NULL;
    *Total              = 0;

    // First, handle NULL or empty string
    if(String == NULL || *String == '\0')
        return Result;
    // count # of property names
    const char *Rover = String;
    for(int N = 1; (Rover=strchr(Rover, SEPARATOR)) != NULL; ++N)
        ++Rover;
    Result  = new TWuiName[N];
    Rover   = String;
    char    Buffer[TWuiName::MAXNAME+1];
    strcpy(Buffer, "p_");
    for(int Index=0; String[0] != '\0'; ++Index)
        {
DEBUG_TRACE("--------->Index=%d, String='%s'", Index, String);
        ASSERT(Index < N);
        Rover   = strchr(String, SEPARATOR);
        if(Rover == NULL)
            Rover   = String + strlen(String);
        size_t  Length  = Rover-String;
        if(Length <= TWuiName::MAXNAME-2)
            {
            StringCopy(Buffer+2, String, Length+1);
            if(TWuiName::LegalName(Buffer) == TRUE)
                {
DEBUG_TRACE("--------->Result[%d] = '%s'", Index, Buffer);
                Result[Index]   = Buffer;
                String  = Rover;
                if(String[0] != '\0')
                    ++String;
                continue;
                }
            }
        // else, not a legal name
        delete[] Result;
        *Total  = -1;
        return NULL;
        }
    *Total  = N;
    return Result;
    }

// SetMruList - Set cached list to that of the named property.
//              Also enforce order!
void    TWuiOrderedContainer::SetMruList(
                                     const TWuiName Property)
    {   MEMBERASSERT();

char Buffer[TWuiName::MAXNAME];
Property.GetName(Buffer);
DEBUG_TRACE("-->SetMruList('%s')", Buffer);

    // Remember which property is being cached
    MruProperty = Property;

    // clear out previous MRU cache
    if(MruList != NULL)
        MruList->RemoveAll();

    // for each child with this property, add to list in sorted order
    TWuiObject     *Child;
    AWuiProperty  *Attribute;
    // call parent GetChild(), avoid infinite recursion
    for(int i=0; (Child=TWuiContainer::GetChild(i)) != NULL; ++i)
        {
DEBUG_TRACE("&&&&[%d]SetMruList() Child=='%s'", i, Child->Name().GetName(Buffer));
DEBUG_TRACE("&&&& Child->GetAttribute() == %s",
    Child->GetAttribute(Property)?"TRUE":"FALSE");
    Attribute   =   (AWuiProperty *)Child->GetAttribute(Property);
DEBUG_TRACE("%%%%IsA(C_TINTPROPERTY) == %s", Attribute->IsA(C_TINTPROPERTY)?"TRUE":"FALSE");
        if((Attribute=(AWuiProperty *)Child->GetAttribute(Property)) != NULL
            &&  Attribute->IsA(C_TINTPROPERTY)
          )
            {
DEBUG_TRACE("&&&&&& Searching this child.");
            int Value   = Attribute->GetInt();

            // Allocate MRU list if it does not exist
            if(MruList == NULL)
                MruList = new TWuiObjects[Children()->NObjects()];
            // insert at sorted location
            for(int j=0; j < MruList->NObjects(); ++j)
                {
                TWuiObject *OtherChild   = (TWuiObject *)MruList->Get(j);
                ASSERTDATAPTR(OtherChild);
                AWuiProperty *OtherProp = (AWuiProperty *)OtherChild->GetAttribute(Property);
                ASSERT(OtherProp != NULL);
                if(OtherProp->GetInt() >= Value)
                    break;
                }
            MruList->Insert(Child, j);
            }
        }
    }

void    TWuiOrderedContainer::RenumberMru(TWuiPath InputPath)
    {   MEMBERASSERT();
DEBUG_TRACE("RenumberMru(InputPath)");
    if(MruList == NULL)
{
DEBUG_TRACE("     (returns because MruList was NULL)");
        return;
}
DEBUG_TRACE("MruList->NObjects() == %d", MruList->NObjects());
    TWuiPath    Path(InputPath);
    // for each child in cache, if numbered incorrectly, fix it
    for(int j=0; j < MruList->NObjects(); ++j)
        {
DEBUG_TRACE("---------------->[%d]---", j);
        TWuiObject *Child   = (TWuiObject *)MruList->Get(j);
        AWuiProperty *Prop  = Child->GetProperty(MruProperty);
        ASSERT(Prop != NULL);
        // if has incorrect value, force it
        if(Prop->GetInt() != j)
            {
            char    Buffer[16];
            sprintf(Buffer, "%d", j);
            Path.Push(Child->Name());
            TWuiArgs    Args(MruProperty);
            Set(Path, &Args, Buffer);
            Path.Pop();
            }
        }
    }


// NewObjectOrder - Handle change to p_ObjectOrder property
void    TWuiOrderedContainer::NewObjectOrder(TWuiPath Path)
    {   MEMBERASSERT();
DEBUG_TRACE("-->NewObjectOrder(%s  %s)", Path.AbsolutePath(),
    Path.RelativePath());
    // delete any previous cached data
    if(PropertyNames != NULL)
        delete[] PropertyNames;
    if(MruList != NULL)
        MruList->RemoveAll();

    // if no new object order field, we are done
    if(ObjectOrder == NULL || *ObjectOrder == '\0')
{
DEBUG_TRACE("Nope, no ObjectOrder field, nothing to do!");
        return;
}

    PropertyNames   = ParseOrderString(ObjectOrder, &NOrdered);
DEBUG_TRACE("<><> After ParseOrderString, NOrdered=%d", NOrdered);
    if(NOrdered < 0)    // if error
        {
        NOrdered    = 0;
        delete[] ObjectOrder;
        ObjectOrder = 0;
        return;
        }
    // for each name in the list, force it into order
    for(int i=0; i < NOrdered; ++i)
        {
        const TWuiName PropName = PropertyNames[i];
{
char Buffer[TWuiName::MAXNAME];
PropName.GetName(Buffer);
DEBUG_TRACE("ooooooo>Propname[%d] = '%s'", i, Buffer);
}
        SetMruList(PropName);
{
char Buffer[TWuiName::MAXNAME];
PropName.GetName(Buffer);
DEBUG_TRACE("ooooooo>Propname[%d] = '%s' after setmrulist", i, Buffer);
}
        RenumberMru(Path);
        }
    
    }


// Set - Handle modifications to "Order" property

int     TWuiOrderedContainer::IsOrdered(const TWuiName Property)
    {
    for(int i=0; i < NOrdered; ++i)
        if(PropertyNames[i] == Property)
            return TRUE;
    return FALSE;
    }

long    TWuiOrderedContainer::Set(TWuiPath Path, TWuiArgs *Args,
                                         const char *Buffer)
    {   MEMBERASSERT();
    if(Path.IsLastNode() && Args->AttributeName() == P_OBJECTORDER)
        {
        // Verify that ObjectOrder string is legal
        int     Status;
        TWuiName    *List = ParseOrderString(Buffer, &Status);
        if(List != NULL)
            delete[] List;
        // if error, do not allow string to be set
        if(Status < 0)
            return WUIMAN_ERR_BAD_ORDER_STRING;
        }
    
    // if one of our child's order fields is being altered,
    // renumber, if necessary
    int         OrderChange = FALSE;
    TWuiObject  *Child;
    TWuiName    Property = Args->AttributeName();
    if(     Path.IsLastNode(1)
         && IsOrdered(Property)
         && ((Child=AWuiObject::GetChild(Path.NextName())) != NULL)
      )
        {
        if(MruProperty != Property)
            SetMruList(Property);
        OrderChange = TRUE;
        }

    long    Result = TWuiContainer::Set(Path, Args, Buffer);
    if(Result < WUIMAN_ERRORS)
        return Result;

    // if one of our child's order fields was altered,
    // renumber, if necessary

    if(OrderChange)
        {
        // Find where child is in old order
        for(int i=0; i < MruList->NObjects(); ++i)
            if(MruList->Get(i) == Child)
                break;
        ASSERT(i < MruList->NObjects());
        // Calculate correct new position
        int NewPosition = atoi(Buffer);
        if(NewPosition < 0)
            NewPosition = 0;
        if(NewPosition > MruList->NObjects())
            NewPosition = MruList->NObjects();
        // Move to correct new position in list
        MruList->Remove(i);
        MruList->Insert(Child, NewPosition);
        RenumberMru(Path);
        }
    // else if the set of order property names was altered
    else if(Args->AttributeName() == P_OBJECTORDER)
            {
            char    Buffer[TWuiPath::MAXPATH];
DEBUG_TRACE("--->Before NewObjectOrder, ObjectOrder='%s'", ObjectOrder);
            NewObjectOrder(Path);
            }
    return Result;
    }

TWuiObject  *TWuiOrderedContainer::GetChild(int Index,
                                 const TWuiName Order) const
    {   MEMBERASSERT();
    if(Order != MruProperty)
        ((TWuiOrderedContainer *)this)->SetMruList(Order);
    if(Index >= 0 && Index < NOrdered)
        return (TWuiObject *)MruList->Get(Index);
    else
        return NULL;
    }

TWuiName    TWuiOrderedContainer::ClassName() const
    {   MEMBERASSERT();
    return C_TORDEREDCONTAINER;
    }
TWuiName    TWuiOrderedContainer::ParentClassName() const
    {   MEMBERASSERT();
    return C_TCONTAINER;
    }

void    TWuiOrderedContainer::Copy(const AWuiObject *Other, int Depth)
    {   MEMBERASSERT();
    TWuiObject::Copy(Other, Depth);
    }

AWuiObject  *TWuiOrderedContainer::Clone(const TWuiName NewName, int Depth)
    {   MEMBERASSERT();
    ASSERT(Children() != NULL);
    ASSERT(Attributes() != NULL);
    AWuiObject *Result = new TWuiOrderedContainer(NewName,
            Attributes()->NObjects(), Children()->NObjects());
    if(Result == NULL)
        return Result;
    else
        {
        Result->Copy(this, Depth);
        return Result;
        }
    }

int     TWuiOrderedContainer::IsA(const TWuiName Class, const TWuiName Method) const
    {   MEMBERASSERT();
    if(Class == C_TORDEREDCONTAINER)
        {
        if(Method == "")
            return TRUE;
        else if(GetAttribute(Method) != NULL)
            return TRUE;
        }
    return TWuiContainer::IsA(Class, Method);
    }

////////////


// TWuiRoot - Special class for root node.
TWuiRoot::TWuiRoot(const TWuiName ObjectName)
    :   TWuiContainer(S_ROOT, 2)
    {   MEMBERASSERT();

#if defined(__SC__)
    char *ThisCompiler = "Symantec C++ v6.1";
#elif defined(__BORLANDC__)
    char *ThisCompiler = "Borland C++ v4.0";
#elif defined(_MSC_VER) 
    char *ThisCompiler = "Microsoft C++ v8.00c";
#else
    char *ThisCompiler = "<Unknown compiler>";
#endif
    Compiler        = StringClone(ThisCompiler);
    WuimanVersion   = StringClone("0.0");
    DefineAttribute(new TWuiStringProperty(P_COMPILER, &Compiler));
    DefineAttribute(new TWuiStringProperty(P_WUIMANVERSION, &WuimanVersion));
    }

TWuiName    TWuiRoot::ClassName() const
    {   MEMBERASSERT();
    return C_TROOT;
    }
TWuiName    TWuiRoot::ParentClassName() const
    {   MEMBERASSERT();
    return C_TCONTAINER;
    }

////////////////////////////////////////////////////////
// AWuiModifiableAttribute - code common to all attributes.

AWuiModifiableAttribute::AWuiModifiableAttribute(
                const TWuiName ObjectName, int InitialFlags)
    :   AWuiAttribute(ObjectName)
    {   MEMBERASSERT();
    SetFlags(InitialFlags);
    }
void    AWuiModifiableAttribute::BindAddress(void *Address)
    {   MEMBERASSERT();
    if(Address == 0)
        ClearFlags(INDIRECT);
    else
        SetFlags(INDIRECT);
    }

TWuiObjects *AWuiModifiableAttribute::Attributes() const
    {   MEMBERASSERT();
    return PropertyProperties;
    }

#if 0
long    AWuiModifiableAttribute::Load(TWuiPath Path,
                                          const char *Value)
    {   MEMBERASSERT();
    if(Value != NULL)
        {
        TWuiArgs    Args(Name());
        return Set(Path, &Args, Value);
        }
    else
        return 0;
    }
#endif

// Save() - attribute save function.
//          
long    AWuiModifiableAttribute::Save(TWuiPath Path,
                                        int Depth) const
    {   MEMBERASSERT();
DEBUG_TRACE(">AWuiModifiableAttribute::Save< '%s', '%s', %d",
    Path.AbsolutePath(), Path.RelativePath(), Depth);

    // if not persistent, we don't save
    if(GetFlags(PERSISTENT) == 0)
        return FALSE;
    long    Result = -1;
    char    SmallBuffer[32];
    // Get value to save by calling derived class function
    // (it will either create string version in buffer we
    //  supply, or return pointer to a string value)
    char *Value = SmallBuffer;
    Result  = AttributeValue(&Value);
    if(Result < WUIMAN_ERRORS)
        DEBUG_FATAL("Can't happen: AttributeValue() failed");
    TCharBuffer Temp(TWuiName::MAXNAME + 16 +strlen(Value));
    char  ThisClass[TWuiName::MAXNAME];
    sprintf(Temp, "%s,%s", ClassName().GetName(ThisClass),
                                                 Value);
    char    AttrName[TWuiName::MAXNAME+1];
    AttrName[0] = '.';
    Name().GetName(AttrName+1);
    char    Path2[256];
    Path.Ascend();
    Path.CopyRelativePath(Path2);
    Path.Descend();
    
#if 0
    DEBUG_TRACE("[AWuiModifiableAttribute::Save] "
        "DB->'%s',attr='%s',val='%s'",
        Path2, AttrName, (char *)Temp);
#endif
    if(!Database->WriteItem(Path2,
                        AttrName, (char *)Temp))
        return WUIMAN_ERR_WRITE_FAILED;
    else    // else, save any child attributes
        return AWuiObject::Save(Path, Depth);
    }

long AWuiModifiableAttribute::Set(TWuiPath Path,
                         TWuiArgs *Args, const char *Buffer)
    {   MEMBERASSERT();
    long    Result = -1;
    if(AWuiObject::TraceFlag)
        Trace("AWuiModifiableAttribute::Set", Path, Args);
    if(Args->AttributeName() != Name())
        return AWuiAttribute::Set(Path, Args, Buffer);

    // we should not get called unless "set" was successful
    // Since "set" succeeded, turn off INHERITED bit
    ClearFlags(INHERITED);

    // if PERSISTENT bit is set, save new value to database
    if(GetFlags(PERSISTENT))
        return Save(Path, 1);
    return Result;
    }

long AWuiModifiableAttribute::Get(TWuiPath Path, TWuiArgs *Args,
      char *Buffer, size_t MaxLength)
    {   MEMBERASSERT();
    if(AWuiObject::TraceFlag)
        Trace("AWuiModifiableAttribute::Get", Path, Args);
    // if caller is asking for custom editing
    if(Args->AttributeName() == M_EDITATTR)
        return FALSE;

    // else if we got called as a node under "/Classes"...
    /*else */if(Args->AttributeName() != Name())
        return AWuiObject::Get(Path, Args, Buffer, MaxLength);
    // else this is just a "normal" attribute get value call
    char    SmallBuffer[64];
    SmallBuffer[0]  = '\0';
    // Call derived class to obtain value as string
    char   *Value   = SmallBuffer;
    long    Result  = AttributeValue(&Value);
    if(Buffer != NULL)
        {
        if(Value == NULL)
            Value   = "";
        size_t  ValueLength = strlen(Value) + 1;
        if(ValueLength > MaxLength)
            {
            ValueLength = MaxLength;
            Result      = WUIMAN_ERR_ATTR_TOO_LONG;
            }
        memcpy(Buffer, Value, ValueLength);
        }
    return Result;
    }

///////////////////////////////
// class AWuiProperty

AWuiProperty::AWuiProperty(
                const TWuiName ObjectName, int InitialFlags)
    :   AWuiModifiableAttribute(ObjectName, InitialFlags)
    {   MEMBERASSERT();
    EnforceName('p');
    }


int     AWuiProperty::IsA(const TWuiName Class, const TWuiName Method) const
    {   MEMBERASSERT();
    if(Class == C_TPROPERTY)
        {
        if(Method == "")
            return TRUE;
        else if(GetAttribute(Method) != NULL)
            return TRUE;
        else
            return FALSE;
        }
    else
        return AWuiModifiableAttribute::IsA(Class, Method);
    }

int     AWuiProperty::GetInt() const
    {   MEMBERASSERT();
    char    SmallBuffer[32];
    char   *Address = SmallBuffer;
    AttributeValue(&Address);
    return atoi(Address);
    }

long    AWuiProperty::GetLong() const
    {   MEMBERASSERT();
    char    SmallBuffer[32];
    char   *Address = SmallBuffer;
    AttributeValue(&Address);
    return atol(Address);
    }

// GetString - use virtual AttributeValue to obtain string from
//             any type of property
size_t  AWuiProperty::GetString(char *To, size_t MaxLength) const
    {   MEMBERASSERT();
    if(To != NULL && MaxLength > 0)
        {
        ASSERTDATAPTR(To);
        char    SmallBuffer[32];
        char   *Address = SmallBuffer;
        AttributeValue(&Address);
        StringCopy(To, Address, MaxLength);
        return strlen(To);
        }
    else
        return 0;
    }

///////////////////////////////
// class TWuiEvent

TWuiEvent::TWuiEvent(
                const TWuiName ObjectName, int InitialFlags)
    :   AWuiModifiableAttribute(ObjectName, InitialFlags)
    {   MEMBERASSERT();
    EnforceName('e');
    }

long    TWuiEvent::AttributeValue(char **Buffer) const
    {   MEMBERASSERT();
    Name().GetName(*Buffer);
    strcat(*Buffer, " (event)");
    return (long)strlen(*Buffer);
    }

///////////////////////////////////////////////////
// class TWuiIntProperty

AWuiObject *TWuiIntProperty::ClassConstructor(const TWuiName Name)
    {
    return new TWuiIntProperty(Name);
    }

TWuiIntProperty::TWuiIntProperty(const TWuiName Name,
                                               int *Address)
    :   AWuiProperty(Name, PERSISTENT|INHERITED|INHERITABLE)
    {   MEMBERASSERT();
    if(Address != NULL)
        {
        IntAddress  = Address;
        SetFlags(INDIRECT);
        }
    else
        IntValue    = 0;
    }

// Internal function for obtaining address of data
inline
int     *TWuiIntProperty::GetAddress() const
    {   MEMBERASSERT();
    if(GetFlags(INDIRECT))
        return IntAddress;
    else
        return (int *)&IntValue;
    }


void    TWuiIntProperty::Copy(const AWuiObject *Other1, int Depth)
    {   MEMBERASSERT();
    AWuiProperty::Copy(Other1, Depth);
    const TWuiIntProperty *Other = (TWuiIntProperty *)Other1;
    int *Address = GetAddress();
    const int *OtherAddress = Other->GetAddress();
    if(Address && OtherAddress)
        *Address    = *OtherAddress;
    }

AWuiObject *TWuiIntProperty::Clone(const TWuiName NewName, int Depth)
    {   MEMBERASSERT();
    AWuiObject *Result = new TWuiIntProperty(NewName);
    if(Result == NULL)
        return Result;
    else
        {
        Result->Copy(this, Depth);
        return Result;
        }
    }

long    TWuiIntProperty::AttributeValue(char **Buffer) const
    {   MEMBERASSERT();
    ASSERT(Buffer != NULL);
    ASSERT(*Buffer != NULL);
    const int *Address = GetAddress();
    sprintf(*Buffer, "%d", *Address);
    return *Address;
    }

long    TWuiIntProperty::Load(TWuiPath /*Path*/, const char *Value)
    {   MEMBERASSERT();
    return SetValue(Value);
    }

long    TWuiIntProperty::GetLong() const
    {   MEMBERASSERT();
    return *GetAddress();
    }
int     TWuiIntProperty::GetInt() const
    {   MEMBERASSERT();
    return *GetAddress();
    }

void    TWuiIntProperty::BindAddress(void *Address)
    {   MEMBERASSERT();
    AWuiModifiableAttribute::BindAddress(Address);
    if(Address != 0)
        IntAddress  = (int *)Address;
    else
        IntValue    = 0;
    }

TWuiName TWuiIntProperty::ClassName() const
    {   MEMBERASSERT();
    return C_TINTPROPERTY;
    }
TWuiName TWuiIntProperty::ParentClassName() const
    {   MEMBERASSERT();
    return C_TPROPERTY;
    }
int     TWuiIntProperty::IsA(const TWuiName Class,
                                const TWuiName Method) const
    {   MEMBERASSERT();
    if(Class == C_TINTPROPERTY)
        {
        if(Method == "")
            return TRUE;
        else if(GetAttribute(Method) != NULL)
            return TRUE;
        else
            return FALSE;
        }
    else
        return AWuiProperty::IsA(Class, Method);
    }

long    TWuiIntProperty::SetValue(const char *Value)
    {   MEMBERASSERT();
    char   *End;
    long    Number = strtol(Value, &End, 0);
    if(End == Value || *End != '\0')
        {
        DEBUG_ERROR("Not a legal value for an int"
                " attribute: '%s'", Value);
        return WUIMAN_ERR_SET_BAD_INTEGRAL;
        }
    if(Number > INT_MAX || Number < INT_MIN)
        {
        DEBUG_ERROR("Integer attribute exceeds integer"
                                     " range: '%s'", Value);
        return WUIMAN_ERR_SET_INT_OVERFLOW;
        }
    
    *GetAddress()   = int(Number);
    return Number;
    }

long    TWuiIntProperty::Set(TWuiPath Path, TWuiArgs *Args,
                                          const char *Value)
    {   MEMBERASSERT();
    if(AWuiObject::TraceFlag)
        Trace("TWuiIntProperty::Set", Path, Args);
    if(Args && Args->AttributeName() != Name())
        return AWuiModifiableAttribute::Set(Path, Args, Value); 

    long    Result = SetValue(Value);
    if(Result <= WUIMAN_ERRORS)
        return Result;
    else
        return AWuiProperty::Set(Path, Args, Value);
    }


///////////////////////////////////////////////////
// class TWuiLongProperty

AWuiObject *TWuiLongProperty::ClassConstructor(const TWuiName Name)
    {
    return new TWuiLongProperty(Name);
    }

TWuiLongProperty::TWuiLongProperty(const TWuiName Name,
                                              long *Address)
    :   AWuiProperty(Name, PERSISTENT|INHERITED|INHERITABLE)
    {   MEMBERASSERT();
    if(Address)
        {
        SetFlags(INDIRECT);
        LongAddress = Address;
        }
    else
        LongValue   = 0;
    }

// Internal function for obtaining address of data
inline
long   *TWuiLongProperty::GetAddress() const
    {   MEMBERASSERT();
    if(GetFlags(INDIRECT))
        return LongAddress;
    else
        return (long *)&LongValue;
    }



void    TWuiLongProperty::Copy(const AWuiObject *Other1, int Depth)
    {   MEMBERASSERT();
    AWuiProperty::Copy(Other1, Depth);
    const TWuiLongProperty *Other = (TWuiLongProperty *)Other1;
    long *Address = GetAddress();
    const long *OtherAddress = Other->GetAddress();
    if(Address && OtherAddress)
        *Address    = *OtherAddress;
    }

AWuiObject *TWuiLongProperty::Clone(const TWuiName NewName, int Depth)
    {   MEMBERASSERT();
    AWuiObject *Result = new TWuiLongProperty(NewName);
    if(Result == NULL)
        return Result;
    else
        {
        Result->Copy(this, Depth);
        return Result;
        }
    }

long    TWuiLongProperty::Load(TWuiPath /*Path*/, const char *Value)
    {   MEMBERASSERT();
    return SetValue(Value);
    }

long    TWuiLongProperty::GetLong() const
    {   MEMBERASSERT();
    return *GetAddress();
    }
int     TWuiLongProperty::GetInt() const
    {   MEMBERASSERT();
    return int(*GetAddress());
    }

long    TWuiLongProperty::AttributeValue(char **Buffer) const
    {   MEMBERASSERT();
    ASSERT(Buffer != NULL);
    ASSERT(*Buffer != NULL);
    const long *Address = GetAddress();
    sprintf(*Buffer, "%ld", *Address);
    return *Address;
    }

void    TWuiLongProperty::BindAddress(void *Address)
    {   MEMBERASSERT();
    AWuiProperty::BindAddress(Address);
    if(Address != 0)
        LongAddress  = (long *)Address;
    else
        LongValue    = 0;
    }

TWuiName TWuiLongProperty::ClassName() const
    {   MEMBERASSERT();
    return C_TLONGPROPERTY;
    }
TWuiName TWuiLongProperty::ParentClassName() const
    {   MEMBERASSERT();
    return C_TPROPERTY;
    }

int     TWuiLongProperty::IsA(const TWuiName Class,
                                const TWuiName Method) const
    {   MEMBERASSERT();
    if(Class == C_TLONGPROPERTY)
        {
        if(Method == "")
            return TRUE;
        else if(GetAttribute(Method) != NULL)
            return TRUE;
        else
            return FALSE;
        }
    else
        return AWuiProperty::IsA(Class, Method);
    }

long    TWuiLongProperty::SetValue(const char *Value)
    {   MEMBERASSERT();
    char   *End;
    long    Number = strtol(Value, &End, 0);
    if(End == Value || *End != '\0')
        {
        DEBUG_ERROR("Not a legal value for an long"
                " attribute: '%s'", Value);
        return WUIMAN_ERR_SET_BAD_INTEGRAL;
        }
    
    *GetAddress()   = Number;
    if(Number <= WUIMAN_ERRORS)
        return TRUE;
    else
        return Number;
    }

long    TWuiLongProperty::Set(TWuiPath Path, TWuiArgs *Args,
                                          const char *Value)
    {   MEMBERASSERT();
    if(AWuiObject::TraceFlag)
        Trace("TWuiLongAttribute::Set", Path, Args);
    if(Args && Args->AttributeName() != Name())
        return AWuiProperty::Set(Path, Args, Value); 

    long Result = SetValue(Value);
    if(Result <= WUIMAN_ERRORS)
        return Result;
    else
        return AWuiProperty::Set(Path, Args, Value);
    }


#ifdef WUI_FLOAT

///////////////////////////////////////////////////
// class TWuiDoubleProperty


AWuiObject *TWuiDoubleProperty::ClassConstructor(const TWuiName Name)
    {
    return new TWuiDoubleProperty(Name);
    }

TWuiDoubleProperty::TWuiDoubleProperty(const TWuiName Name,
                                              double *Address)
    :   AWuiProperty(Name, PERSISTENT|INHERITED|INHERITABLE)
    {   MEMBERASSERT();
    if(Address)
        {
        SetFlags(INDIRECT);
        DoubleAddress = Address;
        }
    else
        DoubleValue   = 0;
    }

void    TWuiDoubleProperty::Copy(const AWuiObject *Other1, int Depth)
    {   MEMBERASSERT();
    AWuiProperty::Copy(Other1, Depth);
    const TWuiDoubleProperty *Other = (TWuiDoubleProperty *)Other1;
    double *Address;
    if(GetFlags(INDIRECT))
        Address = DoubleAddress;
    else
        Address = &DoubleValue;
    const double *OtherAddress;
    if(Other->GetFlags(INDIRECT))
        OtherAddress = Other->DoubleAddress;
    else
        OtherAddress = &Other->DoubleValue;
    if(Address && OtherAddress)
        *Address    = *OtherAddress;
    }

AWuiObject *TWuiDoubleProperty::Clone(const TWuiName NewName, int Depth)
    {   MEMBERASSERT();
    AWuiObject *Result = new TWuiDoubleProperty(NewName);
    if(Result == NULL)
        return Result;
    else
        {
        Result->Copy(this, Depth);
        return Result;
        }
    }

long    TWuiDoubleProperty::AttributeValue(char **Buffer) const
    {   MEMBERASSERT();
    ASSERT(Buffer != NULL);
    ASSERT(*Buffer != NULL);
    const double *Address;
    if(GetFlags(INDIRECT))
        Address = DoubleAddress;
    else
        Address = &DoubleValue;
    sprintf(*Buffer, "%g", *Address);
    return *Address;
    }

void    TWuiDoubleProperty::BindAddress(void *Address)
    {   MEMBERASSERT();
    AWuiProperty::BindAddress(Address);
    if(Address != 0)
        DoubleAddress  = (double *)Address;
    else
        DoubleValue    = 0;
    }

long    TWuiDoubleProperty::Load(const TWuiName &Path, const char *Value)
    {   MEMBERASSERT();
    return SetValue(Value);
    }


TWuiName TWuiDoubleProperty::ClassName() const
    {   MEMBERASSERT();
    return C_TDOUBLEPROPERTY;
    }
TWuiName TWuiDoubleProperty::ParentClassName() const
    {   MEMBERASSERT();
    return C_TPROPERTY;
    }

int     TWuiDoubleProperty::IsA(const TWuiName Class,
                                const TWuiName Method) const
    {   MEMBERASSERT();
    if(Class == C_TDOUBLEPROPERTY)
        {
        if(Method == "")
            return TRUE;
        else if(GetAttribute(Method) != NULL)
            return TRUE;
        else
            return FALSE;
        }
    else
        return AWuiProperty::IsA(Class, Method);
    }

long    TWuiDoubleProperty::SetValue(const char *Value)
    {   MEMBERASSERT();
    char   *End;
    double    Number = strtod(Value, &End);
    if(End == Value || *End != '\0')
        {
        DEBUG_ERROR("Not a legal value for a double"
                " attribute: '%s'", Value);
        return WUIMAN_ERR_SET_BAD_INTEGRAL;
        }

    if(GetFlags(INDIRECT))
        *DoubleAddress = Number;
    else
        DoubleValue    = Number;
    return TRUE;
    }

long    TWuiDoubleProperty::Set(TWuiPath Path, TWuiArgs *Args,
                                          const char *Value)
    {   MEMBERASSERT();
    if(AWuiObject::TraceFlag)
        Trace("TWuiDoubleProperty::Set", Path, Args);
    if(Args && Args->AttributeName() != Name())
        return AWuiProperty::Set(Path, Args, Value); 
    long Result = SetValue(Value);
    if(Result <= WUIMAN_ERRORS)
        return Result;
    else
        return AWuiProperty::Set(Path, Args, Value);
    }

#endif  // WUI_FLOAT


///////////////////////////////////////////////////
// String attributes

AWuiObject *TWuiStringProperty::ClassConstructor(const TWuiName Name)
    {
    return new TWuiStringProperty(Name);
    }

TWuiStringProperty::TWuiStringProperty(const TWuiName Name,
                                             char **Address)
    :   AWuiProperty(Name)
    {   MEMBERASSERT();
    if(Address != NULL)
        {
        SetFlags(INDIRECT);
        StringAddress   = Address;
        }
    else
        StringValue = 0;
    }


TWuiStringProperty::~TWuiStringProperty()
    {   MEMBERASSERT();
    char    **Address = &StringValue;
    if(GetFlags(INDIRECT))
        Address = StringAddress;
    if(*Address != NULL)
        {
        delete[] *Address;
        *Address    = NULL;
        }
    }
void    TWuiStringProperty::Copy(const AWuiObject *Other1, int Depth)
    {   MEMBERASSERT();
    AWuiProperty::Copy(Other1, Depth);
    const TWuiStringProperty *Other = (TWuiStringProperty *)Other1;
    char   **Address;
    if(GetFlags(INDIRECT))
        Address = StringAddress;
    else
        Address = &StringValue;
    char *const*OtherAddress;
    if(Other->GetFlags(INDIRECT))
        OtherAddress = Other->StringAddress;
    else
        OtherAddress = &Other->StringValue;
    if(Address && OtherAddress)
        {
        if(*Address != NULL)
            {
            delete[] *Address;
            *Address = 0;
            }
        if(*OtherAddress)
            *Address    = StringClone(*OtherAddress);
        else
            *Address    = NULL;
        }
    }
AWuiObject *TWuiStringProperty::Clone(const TWuiName Name, int Depth)
    {   MEMBERASSERT();
    AWuiObject *Result = new TWuiStringProperty(Name);
    if(Result == NULL)
        return Result;
    else
        {
        Result->Copy(this, Depth);
        return Result;
        }
    }
void    TWuiStringProperty::BindAddress(void *Address)
    {   MEMBERASSERT();
DEBUG_TRACE("!!!!!!!!!!!!!!BindAddress called!" );
    if(!GetFlags(INDIRECT))
        {
        if(StringValue != 0)
            {
            delete[] StringValue;
            StringValue = 0;
            }
        }
    AWuiProperty::BindAddress(Address);
    if(Address != 0)
        StringAddress  = (char **)Address;
    else
        StringValue    = 0;
    }


long    TWuiStringProperty::AttributeValue(char **Buffer) const
    {   MEMBERASSERT();
    ASSERT(Buffer != NULL);
    ASSERT(*Buffer != NULL);
    char **Address;
    if(GetFlags(INDIRECT))
        Address = (char **)StringAddress;
    else
        Address = (char **)&StringValue;
    // NULL string is treated as empty string
    if(*Address == NULL)
        **Buffer    = '\0';
    else
        *Buffer = *Address;
    return TRUE;
    }

TWuiName TWuiStringProperty::ClassName() const
    {   MEMBERASSERT();
    return C_TSTRINGPROPERTY;
    }

TWuiName TWuiStringProperty::ParentClassName() const
    {   MEMBERASSERT();
    return C_TPROPERTY;
    }

int     TWuiStringProperty::IsA(const TWuiName Class,
                                const TWuiName Method) const
    {   MEMBERASSERT();
    if(Class == C_TSTRINGPROPERTY)
        {
        if(Method == "")
            return TRUE;
        else if(GetAttribute(Method) != NULL)
            return TRUE;
        else
            return FALSE;
        }
    else
        return AWuiProperty::IsA(Class, Method);
    }

long    TWuiStringProperty::Load(TWuiPath /*Path*/, const char *Value)
    {   MEMBERASSERT();
    return SetValue(Value);
    }
long    TWuiStringProperty::SetValue(const char *Value)
    {   MEMBERASSERT();
    char    *Result = StringClone(Value);
    if(Result == NULL)
        return WUIMAN_ERR_NEW_FAILED;
    char    **Address = GetFlags(INDIRECT)?StringAddress:&StringValue;
    ASSERT(Address != NULL);
    if(*Address)
        {
        delete[] *Address;
        *Address    = 0;
        }
    *Address    = Value ? StringClone(Value) : 0;
    return TRUE;
    }

// Set - delete our current value, make a copy of the new string.
long    TWuiStringProperty::Set(TWuiPath Path,
                                        TWuiArgs *Args,
                                         const char *Buffer)
    {   MEMBERASSERT();
    if(AWuiObject::TraceFlag)
        Trace("TWuiStringProperty::Set", Path, Args);
    long    Result  = SetValue(Buffer);
    if(Result <= WUIMAN_ERRORS)
        return Result;
    else
        return AWuiModifiableAttribute::Set(Path, Args, Buffer); 
    }


TWuiMethod::TWuiMethod(const TWuiName ObjectName)
    :   AWuiAttribute(ObjectName)
    {   MEMBERASSERT();
    EnforceName('m');
    }

void    TWuiMethod::Copy(const AWuiObject *Other, int Depth)
    {   MEMBERASSERT();
    AWuiAttribute::Copy(Other, Depth);
    }

AWuiObject *TWuiMethod::Clone(const TWuiName ObjectName, int Depth)
    {   MEMBERASSERT();
    AWuiObject *Result = new TWuiMethod(ObjectName);
    if(Result == NULL)
        return Result;
    else
        {
        Result->Copy(this, Depth);
        return Result;
        }
    }

long TWuiMethod::AttributeValue(char **Buffer) const
    {   MEMBERASSERT();
    ASSERT(Buffer != NULL);
    ASSERT(*Buffer != NULL);
    **Buffer    = '\0';
    return 0;
    }

TWuiName    TWuiMethod::ClassName() const
    {   MEMBERASSERT();
    return C_TMETHOD;
    }
TWuiName    TWuiMethod::ParentClassName() const
    {   MEMBERASSERT();
    return C_TATTRIBUTE;
    }

AWuiAttribute::AWuiAttribute(const TWuiName ObjectName)
    :   AWuiObject(ObjectName)
    {   MEMBERASSERT();
    }

void    AWuiAttribute::EnforceName(char Prefix)
    {   MEMBERASSERT();
    char    ThisName[TWuiName::MAXNAME+1];
    Name().GetName(ThisName);
    if(ThisName[0] != Prefix || ThisName[1] != '_')
        {
        memmove(ThisName+2, ThisName, TWuiName::MAXNAME-2);
        ThisName[0] = Prefix;
        ThisName[1] = '_';
        ThisName[TWuiName::MAXNAME] = '\0';
        Rename(ThisName);
        }
    }

TWuiName    AWuiAttribute::ClassName() const
    {   MEMBERASSERT();
    return C_TATTRIBUTE;
    }

TWuiName    AWuiAttribute::ParentClassName() const
    {   MEMBERASSERT();
    return 0;
    }

int     AWuiAttribute::IsA(const TWuiName Class, const TWuiName Method) const
    {   MEMBERASSERT();
    if(Class == C_TATTRIBUTE)
        {
        if(Method == "")
            return TRUE;
        else if(GetAttribute(Method) != NULL)
            return TRUE;
        }
    // we are top-most WUI class, so no parent to resort to
    return FALSE;
    }


///////////////////////////////////////////////////
// Internal Bit attributes

TWuiIntPropertyBit::TWuiIntPropertyBit(
            const TWuiName Name)
    :   AWuiProperty(Name)
    {   MEMBERASSERT();
    SetFlags(REFERENCE|PERSISTENT);
    }

void    TWuiIntPropertyBit::Copy(const AWuiObject * /*Other*/, int /*Depth*/)
    {   MEMBERASSERT();
    // do nothing, copying of bits is handled by
    // AWuiObject::Copy()
    return;
    }
AWuiObject *TWuiIntPropertyBit::Clone(
        const TWuiName /*Name*/, int /*Depth*/)
    {   MEMBERASSERT();
    ASSERT(FALSE);  // Should never be called!
    return NULL;
    }


TWuiName TWuiIntPropertyBit::ClassName() const
    {   MEMBERASSERT();
    return C_TINTERNALBIT;
    }
TWuiName TWuiIntPropertyBit::ParentClassName() const
    {   MEMBERASSERT();
    return C_TINTPROPERTY;
    }
long    TWuiIntPropertyBit::AttributeValue(char **Buffer) const
    {   MEMBERASSERT();
    ASSERT(Buffer != NULL);
    ASSERT(*Buffer != NULL);

    int On = GetFlags(0x8000);
    if(On != 0)
        strcpy(*Buffer, "1");
    else
        strcpy(*Buffer, "0");
    return On != 0;
    }

long    TWuiIntPropertyBit::Set(TWuiPath Path, TWuiArgs *Args,
                                          const char *Value)
    {   MEMBERASSERT();
    if(AWuiObject::TraceFlag)
        Trace("TWuiIntPropertyBit::Set", Path, Args);
    int On   = 0;
    if(Value != NULL)
        On  = (atoi(Value) != 0);
    // very tacky hack. essentially reserving topmost flag bit
    // to communicate with our AttributeValue() function across
    // the call to AWuiModifiableAttribute::Set().  Ugh.
    SetFlags(0x8000, On);
    return AWuiProperty::Set(Path, Args, Value);
    }

